
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include "pz.h"

#define DIV  16
#define MULT 17
#define SUBT 18
#define ADD  19
#define SIN 20
#define COS 21
#define TAN 22
#define EQL  23
#define DP   24
#define LN 25
#define LOG 26
#define POW 27
#define ARCS 28
#define ARCC 29
#define ARCT 30
#define POS 31
#define EB 14
#define PIB 15
#define PI 3.1415926535
#define E 2.7182818285



#define N_BUTTONS (28 - 1)
struct _buttons {
	char *ch;
	unsigned char math;
} buttons[] = {
	{"7", 7}, {"8", 8}, {"9", 9}, {"/", DIV}, {"sin", SIN}, {"arcs", ARCS}, {"ln", LN},
	{"4", 4}, {"5", 5}, {"6", 6}, {"*", MULT}, {"cos", COS}, {"arcc", ARCC}, {"log", LOG},
	{"1", 1}, {"2", 2}, {"3", 3}, {"-", SUBT}, {"tan", TAN}, {"arct", ARCT}, {"^", POW},
	{"0", 0}, {".", DP},{"=", EQL},{"+",ADD}, {"pi", PIB}, {"+-", POS}, {"e", EB},
};

static PzModule *module;
static long double number, lastnumber;
static signed char e;
static unsigned char op;
static int cur_button;

static void draw_button(PzWidget *wid, ttk_surface srf, int row, int col)
{
#define XOFF (wid->w / 10)
#define YOFF (wid->h / 4)
#define SECW ((wid->w - XOFF*2)/6) 
#define SECH ((wid->h - YOFF - wid->h/10)/4) 
#define BPAD ((SECW - ttk_text_width(ttk_textfont, "10"))/4)
#define BW (SECW - BPAD)
#define BH (SECH - BPAD)
	char *text[] = {"button.selected.fg", "button.default.fg",
		"button.selected.bg", "button.default.bg",
		"button.selected.border", "button.default.border"};
	int pos = (row * 7) + col;
	unsigned char ap = !(pos == cur_button);
	int x = 10 + col*(BW + BPAD);
	int y = 5 + (YOFF + BPAD/2) + row*(BH + BPAD);

	ttk_ap_fillrect(srf, ttk_ap_get(text[ap+2]), x+1, y+1, x+BW-1, y+BH-1);
	ttk_ap_rect(srf, ttk_ap_get(text[ap+4]), x, y, x + BW, y + BH);
	ttk_text(srf, ttk_textfont, x+(BW-ttk_text_width(ttk_textfont, "10"))/2,
			y+(BH-ttk_text_height(ttk_textfont))/2,
			ttk_ap_getx(text[ap])->color, buttons[pos].ch);
}

static void calculator_draw(PzWidget *wid, ttk_surface srf)
{
	int row, col;
	int x = XOFF + BPAD/2;
	int y = wid->h/8;
	int xe = x + (wid->w - XOFF*2) - BPAD;
	char s[228];
	char v[128];
	double a;
	long int b, t;
	
	
	t = floor(number);
	
	
	a = number - t;
	a = fabsl(a);
	a *= 10;
	
	snprintf(s, 228, "%d", t);
	snprintf(v, 128, "%f", a);
	strcat(s,".");
	strcat(s,v);
	ttk_rect(srf, x, y, xe, y+BH, ttk_ap_getx("window.fg")->color);
	ttk_text(srf, ttk_textfont, xe - ttk_text_width(ttk_textfont, s) - 1,
			y + (BH - ttk_text_height(ttk_textfont))/2,
			ttk_ap_getx("window.fg")->color, s);
	
	for (row = 0; row < 4; row++)
		for (col = 0; col < 7; col++)
			draw_button(wid, srf, row, col);
}

static double lpow(double x, signed char p)
{
	double i = x;
	if (p > 0) while (--p > 0) x *= i;
	else if (p-- < 0) while (p++ < 0) x /= i;
	else return 1;
	return x;
}

static double operation(double a, unsigned char op, double b)
{
	switch (op) {
	case DIV: return a / b;
	case MULT:return a * b;
	case SUBT:return a - b;
	case ADD: return a + b;
	case POW: return pow(a,b);
	}
	return 0;
}

static void calculate(int pos)
{
	struct _buttons *p = &buttons[pos];
	static int stump = 0;
	if (p->math & 16) {
		switch (p->math) {
		case DIV:
		case MULT:
		case SUBT:
		case ADD:
		case POW:
			if (op) number = operation(lastnumber, op, number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case SIN:
			number = sin(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case TAN:
			number = tan(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case COS:
			number = cos(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case LN:
			number = log(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case LOG:
			number = log10(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case ARCS:
			number = asin(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case ARCC:
			number = acos(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case ARCT:
			number = atan(number);
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case POS:
			number *= -1;
			op = p->math;
			lastnumber = number;
			stump = 1;
			e = 0;
			break;
		case EQL:
			number = operation(lastnumber, op, number);
			op = 0;
			stump = 1;
			e = 0;
			break;
		case DP:
			if (!e) --e;
			break;
		}
	}
	else if (p->math == 15) {
			number = 3.1415926535;
			stump = 0;
	}
	else if (p->math == 14) {
			number = 2.7182818285;
			stump = 0;
	}
	else if (!stump) {
		if (!e) number *= 10;
		number += p->math * lpow(10, e);
		if (e < 0) --e;
	}
	else {
		number = p->math;
		stump = 0;
	}
}

#define BWRAP(x,y,z) if (x < y) x = z
#define TWRAP(x,y,z) if (x > z) x = y
#define WRAP(x,y,z) do { BWRAP(x,y,z); TWRAP(x,y,z); } while(0)

static int calculator_events(PzEvent *e)
{
	int ret = TTK_EV_CLICK;
	
	switch(e->type) {
	case PZ_EVENT_BUTTON_DOWN:
		switch (e->arg) {
		case PZ_BUTTON_ACTION:
			calculate(cur_button);
			e->wid->dirty = 1;
			break;

		case PZ_BUTTON_MENU:
			pz_close_window(e->wid->win);
			break;
		
		case PZ_BUTTON_PLAY:
			calculate(EQL);
			break;
			
		
		default:
			ret = TTK_EV_UNUSED;
		}
		break;

	case PZ_EVENT_SCROLL:
		TTK_SCROLLMOD(e->arg, 5);
		cur_button += e->arg;
		WRAP(cur_button, 0, N_BUTTONS);
		e->wid->dirty = 1;
		break;

	default:
		ret = TTK_EV_UNUSED;
		break;
	}
	return ret;
}

static PzWindow *new_calculator()
{
	PzWindow *ret;
	number = lastnumber = 0;
	e = 0;
	cur_button = 0;
	op = 0;

	ret = pz_new_window(_("Scientific Calculator"), PZ_WINDOW_NORMAL);
	pz_add_widget(ret, calculator_draw, calculator_events);

	return pz_finish_window(ret);
}

static void init_calculator()
{
	module = pz_register_module("calc", NULL);
	pz_menu_add_action("/Extras/Utilities/Scientific Calculator", new_calculator);
}

PZ_MOD_INIT(init_calculator)
